#!/usr/bin/env python
#
#"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
# Copyright (C) 2007 Hubert Berezowski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
#
#"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
#   .~SSS' '' ' s'     `~:.  `.SSSSS'''''aS  
#     >Ss''''.'SSSs       ';    >SSS'''.'SS
#   ,'''S', ~`'SSSSSs    `~   ,''''S''  .  
# '''',''   A. `'SSSSSs    ,''''','    `A;.
#'' s'   K. ..N. `'SSSSS'''''''sSs. ~a;.,;   
#''SSSs    E.~  Y   >SSS'''';'.SSSSs. ~a.``      
#`'SSSSSs    Y    ,''''S',' ,. `'SSSSs. ~;>   
#  `'SSSSSs.   ,'''''','    'I;. `'SSSSs. ,  
#;k   `'SSSS''''''''s'  <E;... ~L. `'SSSS''
#df     ,SSS'''''.'SSSs   `X. ``~'E>  >SS''
#     `''''S''''` `'SSSSs   ~A.    ,`'''S''
#
# ADD KEYBOARD SHORTCUTS TO EXAILE



from gettext import gettext as _

import xl.plugins as plugins

import gtk
from gobject import source_remove,io_add_watch
from xl import common

try:
    from Xlib.display import Display
    from Xlib import X 
except ImportError:
    X = None

PLUGIN_NAME = _("Keyboard Shortcuts")
PLUGIN_AUTHORS = ['Hubert Berezowski <hubertb2@wp.pl>']
PLUGIN_VERSION = '0.1.6'
PLUGIN_DESCRIPTION = _(r""" Keyboard Shortcuts via the X Server.
\n\n
Requires Python-Xlib
\n
http://python-xlib.sourceforge.net/""")


PLUGIN_ENABLED = False

b = gtk.Button()
PLUGIN_ICON = b.render_icon(gtk.STOCK_BOLD, gtk.ICON_SIZE_MENU)
b.destroy()
 
CONNS = plugins.SignalContainer()




class XlibKeys:
    def __init__(self):
        '''
        test for settings and connect them with xlib 
        '''


        # actionlist, the keys of the keyb dictionary, sorting order in the configure panel
        self.actionlist = ['play','pause','stop','next Track','previous Track','seek forward','seek backward','volume up','volume down','show osd']

        # load settings or use default
        try:
            self.getsettings()
        except:

            #        [key] [button] [ 1=pressed 0=unpressed       ] [ a function the key will call ]
            #        [   ] [text  ] [ leave this 0                ] [ or a direct exaile command ]
            #         v      v       v       v     v    v   v    v    v        v
            # keyb = 'name':[keyname,keycode,shift,ctrl,alt,mod1,mod2,callback,callback-arguments]
            #         ^       ^      ^       ^     ^    ^   ^    ^    ^        ^        
            # keyb = '*   ':[str    ,int    ,int  ,int ,int,int,int  ,str     ,str               ]

            
            self.keyb = {'play':          ['not defined', 0, 0, 0, 0, 0, 0, 'APP.player.play',''],
                         'stop':          ['not defined', 0, 0, 0, 0, 0, 0, 'APP.player.stop',''],
                         'pause':         ['not defined', 0, 0, 0, 0, 0, 0, 'APP.player.toggle_pause',''],
                         'next Track':    ['not defined', 0, 0, 0, 0, 0, 0, 'APP.player.next',''],
                         'previous Track':['not defined', 0, 0, 0, 0, 0, 0, 'APP.player.previous',''],
                         'seek forward':  ['not defined', 0, 0, 0, 0, 0, 0, 'seekForward','5'],
                         'seek backward': ['not defined', 0, 0, 0, 0, 0, 0, 'seekBack','5'],
                         'volume up':     ['not defined', 0, 0, 0, 0, 0, 0, 'volUp5','2'],
                         'volume down':   ['not defined', 0, 0, 0, 0, 0, 0, 'volDown5','2'],
                         'show osd':      ['not defined', 0, 0, 0, 0, 0, 0, 'APP.show_osd',''],
                         }

        # connect loaded settings with X
        # start listening
        self.listen()

    def ButtonHit(self,event):
        '''##########################################################################
        a mapped button was hit! do specified exaile action
        '''
        # specify actions:
        def seekBack(step):
            APP.player.seek(APP.player.get_position()/1000000000-step)
            APP.show_osd()

        def seekForward(step):
            APP.player.seek(APP.player.get_position()/1000000000+step)
            APP.show_osd()
            
        def volUp5(step):
            APP.volume.set_value(APP.volume.value+step)
            
        def volDown5(step):
            APP.volume.set_value(APP.volume.value-step)
            
        # run the function keyb[callback]([arguments]) if keyb[kecode] == x.event
        for i in self.keyb.values():
            if i[1] == event:
                eval (i[7]+'('+i[8]+')')

    def extras(self,key):
        '''#########################################################################
         called by updatebuttons() in the configure method. will return a gtk widget
         which will be between the label and key-button.
         '''
        def setextra(value):
            self.keyb[key][8]=str(value.value)
            
        if key == 'seek forward':
            container = gtk.HBox(False, 0)
            adjustments = gtk.Adjustment(float(self.keyb[key][8]), 1, 50, 1, 0, 0)
            adjustments.connect("value_changed", setextra)
            zeichen = gtk.Label(" ")        
            label = gtk.Label(_("sec."))        
            steps = gtk.HScale(adjustments)        
            steps.set_value_pos(gtk.POS_LEFT)
            steps.set_usize( 100, 0)
            container.pack_start(zeichen, True, True, 3)        
            container.pack_start(steps, True, True, 3)        
            container.pack_start(label, False, False, 3)
            zeichen.show()
            steps.show()
            label.show()

            return container

        if key == 'seek backward':
            container = gtk.HBox(False, 0)
            adjustments = gtk.Adjustment(float(self.keyb[key][8]), 1, 50, 1, 0, 0)
            adjustments.connect("value_changed", setextra)
            zeichen = gtk.Label("-")                    
            label = gtk.Label(_("sec."))        
            steps = gtk.HScale(adjustments)        
            steps.set_value_pos(gtk.POS_LEFT)
            steps.set_usize( 100, 0)
            container.pack_start(zeichen, True, True, 3)        
            container.pack_start(steps, True, True, 3)        
            container.pack_start(label, False, False, 3)
            #steps.set_inverted(True)
            zeichen.show()
            steps.show()
            label.show()

            return container

        if key == 'volume up':
            zeichen = gtk.Label(" ")                    
            container = gtk.HBox(False, 0)
            adjustments = gtk.Adjustment(float(self.keyb[key][8]), 0, 25, 1, 0, 0)
            adjustments.connect("value_changed", setextra)
            steps = gtk.HScale(adjustments)        
            steps.set_value_pos(gtk.POS_LEFT)
            steps.set_usize( 100, 0)
            label = gtk.Label("%")        
            container.pack_start(zeichen, True, True, 3)        
            container.pack_start(steps, True, True, 3)
            container.pack_start(label, False, False, 3)
            steps.show()
            zeichen.show()
            label.show()
            
            return container

        if key == 'volume down':
            zeichen = gtk.Label("-")            
            container = gtk.HBox(False, 0)
            adjustments = gtk.Adjustment(float(self.keyb[key][8]), 0, 25, 1, 0, 0)
            adjustments.connect("value_changed", setextra)
            steps = gtk.HScale(adjustments)        
            #steps.set_inverted(True)
            steps.set_value_pos(gtk.POS_LEFT)
            steps.set_usize( 100, 0)
            label = gtk.Label("%")                    
            container.pack_start(zeichen, True, True, 3)                    
            container.pack_start(steps, True, True, 3)
            container.pack_start(label, False, False, 3)
            steps.show()
            zeichen.show()
            label.show()
            
            return container


    
##################################################################################
#
# If you want to add your custom shortcut, this is as far as you have to edit this
# file. The Rest is generic stuff that should work with any action. As long as it
# follows following pattern:
#
#     in the __init__ method:
#         a valid entry(see comments) in the keyb dictionary and actionlist
#
#  optional:
#     in the Buttonhit method:
#         a function called by the 7the entry in the dict
#
#     in the extras method:
#         additional config menu items
#
##################################################################################


    def freeKey(self):
        '''#######################################################################
        stop grabbing without a trace (i hope)
        '''

        source_remove(self.listener)
        self.grabKey('free')
        #self.disp.close()


    def grabKey(self,action):
        '''#######################################################################
        Connect / disconnect Keys to grab
        '''
        # Make a xlib Mask for every key in the settings-dict
        for key in self.keyb.keys():
            if self.keyb[key][0] != 'not defined':
                maske = ''
                if self.keyb[key][2] == 1:
                    maske += '|X.ShiftMask'
                if self.keyb[key][3] == 1:
                    maske += '|X.ControlMask'
                if self.keyb[key][4] == 1:
                    maske += '|X.Mod1Mask'
                if self.keyb[key][5] == 1:
                    maske += '|X.Mod2Mask'
                if self.keyb[key][6] == 1:
                    maske += '|X.Mod3Mask'

            
                def checkmask(maske):
                    # make sure the mask is valid
                    if maske != '':
                        amaske = maske.lstrip('|')
                        return eval(amaske)
                    else:
                        return X.AnyModifier
            
                if action == 'grab':
                    # ...and grab it...
                    #print 'grab hit'
                    self.root.grab_key(self.keyb[key][1], checkmask(maske),  1,X.GrabModeAsync, X.GrabModeAsync)
                    #print 'grabbed '+str(self.keyb[key][1])
                elif action == 'free':
                    # ...or ungrab it.
                    self.root.ungrab_key(self.keyb[key][1], checkmask(maske))
                    self.disp.flush()
                    #print 'ungrabbed '+str(self.keyb[key][1])
                
    def listen(self):
        '''##########################################################################
        connect xlib to the dictionary with the keys
        '''


        # select display to listen on
        self.disp = Display()
        self.root = self.disp.screen().root
        # specify event
        self.root.change_attributes(event_mask = X.KeyPressMask)

        self.grabKey('grab')

        #print 'listening'
        def checkKey(arg1,arg2):
            #print arg1,arg2
            event = self.disp.next_event()
            if event.type == X.KeyPress:
                self.ButtonHit(event.detail)
            return True
    
        self.listener = io_add_watch(self.disp, 1 ,checkKey)
        print self.disp.pending_events()


    def savesettings(self):
        '''##########################################################################
        Save the working dictionary a a string
        '''
        #print 'savesettings'
        for k in self.keyb.keys():
            #print self.keyb[k]
            APP.settings.set_str(k, str(self.keyb[k]), plugin=plugins.name(__file__))



        
    def getsettings(self):
        '''##########################################################################
        Get the string and convert it back to the state it was before saving
        '''
        #value = get_str('setting_name_1', default='some string', plugin=plugins.name(__file__))
        # this dictionary will hold the setting
        self.keyb = {}

        for action in self.actionlist:


            # get values, remove garbage and split values into a list
            values = APP.settings.get_str(action, default='some string', plugin=plugins.name(__file__)).strip('[]').rsplit(',')

            # convert values into desired types
            clean = [values[0].lstrip(' ').strip('\''),int(values[1]),int(values[2]),int(values[3]),int(values[4]),int(values[5]),int(values[6]),values[7].lstrip(' ').strip('\''),values[8].lstrip(' ').strip('\'')]

            # finally put all into the dict
            self.keyb[action] = clean
            


    def configdialog(self):
        '''##########################################################################

        '''
        def defineNewKey(widget,data):

            def keypressed(widget, event):
                mod = 0
                # put pressed modifiers into the list and count them
                ####
                if 'GDK_SHIFT_MASK' in event.state.value_names:
                    self.keyb[data][2] = 1
                    mod += 1
                else:
                    self.keyb[data][2] = 0
                    
                if 'GDK_CONTROL_MASK' in event.state.value_names:
                    self.keyb[data][3] = 1
                    mod += 1                    
                else:
                    self.keyb[data][3] = 0

                if 'GDK_MOD1_MASK' in event.state.value_names:
                    self.keyb[data][4] = 1
                    mod += 1
                else:
                    self.keyb[data][4] = 0

                if 'GDK_MOD2_MASK' in event.state.value_names:
                    self.keyb[data][5] = 1
                    mod += 1
                else:
                    self.keyb[data][5] = 0

                if 'GDK_MOD3_MASK' in event.state.value_names:
                    self.keyb[data][6] = 1
                    mod += 1
                else:
                    self.keyb[data][6] = 0

                #### now sort out the exceptions
                # omitt doing an action for modifier key events                
                if event.string == '':
                    print 'mod hit'
                    
                elif mod == 0:
                    print 'no mod used'
                    
                else:
                    # If the user entered a valid keycombination, save the key in the dict
                    # and reload the main-plugin-configuration dialog
                    # so that the label on the button shows the correct key
                    # but first check if the key is already mapped
                    used = ''
                    for kcode in self.keyb:
                        if self.keyb[kcode][1] == event.hardware_keycode:
                            if self.keyb[kcode][0] != self.keyb[data][0]:
                                used = kcode
                            
                    if used == '':
                        self.keyb[data][0] = gtk.gdk.keyval_name(event.keyval)
                        self.keyb[data][1] = event.hardware_keycode
                        pkeyinfo.destroy()
                        updatebuttons()
                    else:
                        useddialog = gtk.MessageDialog(APP.window, gtk.DIALOG_DESTROY_WITH_PARENT,
                                                       gtk.MESSAGE_INFO,
                                                       gtk.BUTTONS_OK,
                                                       _("Key already mapped by %s") % used)
                        useddialog.run()
                        useddialog.destroy()

                
            #######################################################################
#        dialog = gtk.Dialog("Configure Exaile Shortcuts",
#                            APP.window,
#                            gtk.DIALOG_DESTROY_WITH_PARENT,
#                            (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
#                             gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
        

                        
            pkeyinfo = gtk.MessageDialog(APP.window, gtk.DIALOG_DESTROY_WITH_PARENT,
                                         gtk.MESSAGE_INFO, gtk.BUTTONS_CANCEL,
                                         _("Please Press a Key for '%s' while holding one or more modifier Keys (ctrl, shift, alt)") % data)
            
            pkeyinfo.add_events(gtk.gdk.KEY_PRESS_MASK)
            pkeyinfo.connect('key-press-event', keypressed)
            pkeyinfo.run()
            pkeyinfo.destroy()
            
        def clearKey(widget, data):
            # user clicked the 'clear' button
            self.keyb[data][0]='not defined'
            self.keyb[data][1]=0
            self.keyb[data][2]=0
            self.keyb[data][3]=0
            self.keyb[data][4]=0
            self.keyb[data][5]=0                              
            self.keyb[data][6]=0
            updatebuttons()
            
        # packung -------------------------------------------------+
        #    label     key-button             clear-button         |
        #               L  defineNewKey()      L  self.clearKey()  |      
        #                   L  Keypressed()                        |
        #----------------------------------------------------------+


        dialog = gtk.Dialog(_("Configure Exaile Shortcuts"),
                            APP.window,
                            gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                             gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
        
        beschreibung = gtk.Label(_('The shortcuts will be globally accessible in X'))
        dialog.vbox.pack_start(beschreibung, False, False, 0)

        def updatebuttons():
            # remove the refreshbox container holding the old buttons
            #try:
            dialog.vbox.remove(self.refreshbox)

            # create the context for the buttons 
            self.refreshbox = gtk.VBox(True, 0)
            dialog.vbox.pack_start(self.refreshbox, False, False, 5)
            beschreibung.set_padding(5, 3)
            beschreibung.show()
            dialog.set_border_width(5)
            
            
            for key in self.actionlist:

                # create something like this for every key:
                #
                #  +- packung -----------------------------------+
                #  | [ name ] ( [ extra ] )? [ kname ] [ clk ]   |
                #  +---------------------------------------------+
                #
                # and put it in the refresh-vbox:
                #  
                #       +- dialog.vbox -------+
                #       | beschreibung(label) |
                #   --> | +- refreshbox ----+ |
                #       | |  [ packung 1 ]  | |
                #       | |  [ packung 2 ]  | |
                #       | |  etc ...        | |
                #   --> | +-----------------+ |
                #       |        [cancel][ok] |
                #       +---------------------+                
                
                packung = gtk.HBox(False, 0)

                
                knopf = str(self.keyb[key][0])
                if self.keyb[key][2] == 1:
                    knopf += ' + shift'
                if self.keyb[key][3] == 1:
                    knopf += ' + ctrl'
                if self.keyb[key][4] == 1:
                    knopf += ' + mod1'
                if self.keyb[key][5] == 1:
                    knopf += ' + mod2'
                if self.keyb[key][6] == 1:
                    knopf += ' + mod3'

                name  = gtk.Label(key)
                name.set_width_chars(15)
                kname = gtk.Button(knopf)
                clk   = gtk.Button('clear')

                kname.connect("clicked",defineNewKey, key)
                clk.connect("clicked",clearKey, key)
                self.refreshbox.pack_start(packung, False, False, 5)                
                packung.pack_start(name, False, False, 5)

                # check for extra options
                if self.keyb[key][8] != '':
                    extra = self.extras(key)
                    packung.pack_start(extra, False, False,5)
                    extra.show()

                packung.pack_start(kname, True, True, 5)
                packung.pack_start(clk, False, False, 2)
                name.show()
                kname.show()
                clk.show()
                packung.show()
            self.refreshbox.show()

            
               

        # ungrab keys, so that no key is blocked while defining a new shortcut
        self.grabKey('free')
        source_remove(self.listener)

        # create the context for the buttons 
        self.refreshbox = gtk.VBox(True, 0)
        dialog.vbox.pack_start(self.refreshbox, False, False, 5)
        beschreibung.set_padding(5, 3)
        beschreibung.show()
        dialog.set_border_width(5)


        updatebuttons()
        
        # If user clicked 'ok' save settings else reload settings
        dial = dialog.run()
        if dial == gtk.RESPONSE_ACCEPT:
            self.savesettings()
            self.listen()
        elif dial == gtk.RESPONSE_REJECT:
            self.getsettings()
            self.listen()
        dialog.destroy()





def initialize():
    """
        Called when the plugin is enabled
    """
    if not X:
        common.error(APP.window, _('Python-xlib library is not available. '
            'Cannot initialize xkeys plugin.'))
        return False
    APP.window.cuts = XlibKeys()
    return True
 
def destroy():
    """
        Called when the plugin is disabled
    """
    APP.window.cuts.freeKey()

    CONNS.disconnect_all()
 
def configure():
    """
        Called when the user clicks Configure in the Plugin Manager
    """
    APP.window.cuts.configdialog()





#!/usr/bin/env python
import gtk
from gettext import gettext as _
import xl.plugins as plugins

PLUGIN_NAME = _("No Taskbar Entry")
PLUGIN_AUTHORS = ['Jonas Wagner <veers' + chr(32+32) + 'gmx' + '.ch>']
PLUGIN_VERSION = "0.1.2"
PLUGIN_DESCRIPTION = _(r"""Removes exaile from the taskbar""")

PLUGIN_ENABLED = False
PLUGIN_ICON = None
PLUGIN = None

def initialize():
	APP.window.set_property("skip-taskbar-hint", True)
	return True

def destroy():
	APP.window.set_property("skip-taskbar-hint", False)



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

from xl import media, library, common, xlmisc, editor
from gettext import gettext as _, ngettext
import gtk, gtk.glade
import re, os
import playlist as trackslist

TAG_STRINGS = {
    "title":            _('Title'),
    "version":          _('Version'),
    "album":            _('Album'),
    "tracknumber":      _('Track Number'),
    "artist":           _('Artist'),
    "genre":            _('Genre'),
    "performer":        _('Performer'),
    "copyright":        _('Copyright'),
    "license":          _('License'),
    "organization":     _('Organization'),
    "description":      _('Description'),
    "location":         _('Location'),
    "contact":          _('Contact'),
    # TRANSLATORS: International Standard Recording Code
    "isrc":             _('ISRC'),
    "date":             _('Date'),
    "arranger":         _('Arranger'),
    "author":           _('Author'),
    "composer":         _('Composer'),
    "conductor":        _('Conductor'),
    "lyricist":         _('Lyricist'),
    "discnumber":       _('Disc Number'),
    "labelid":          _('Label ID'),
    "part":             _('Part'),
    "website":          _('Website'),
    "language":         _('Language'),
    # TRANSLATORS: Beats per minute
    "bpm":              _('BPM'),
    "albumartist":      _('Album Artist'),
    "originaldate":     _('Original Date'),
    "originalalbum":    _('Original Album'),
    "originalartist":   _('Original Artist'),
    "recordingdate":    _('Recording Date'),
    "encodedby":        _('Encoded By'),
}

TAGNAME, VALUE, OLDVALUE, ERROR, ADDED, REMOVED, EDITED = range(0,7)

# these items are read from the database, so they will need to be copied over
# when creating the songs list (see FIXME below)
DB_COPY_ITEMS = ('modified', '_rating', 'system_rating', 'blacklisted', 'time_added',
    'encoding', 'playcount')

class TrackEditor(object):
    """
        A track properties editor
    """


    def __init__(self, exaile, tracks):
        """
            Initializes the panel 
        """
        self.exaile = exaile
        self.db = exaile.db
        self.xml = gtk.glade.XML('exaile.glade', 'TrackEditorDialog', 'exaile')
        self.dialog = self.xml.get_widget('TrackEditorDialog')
        self.dialog.set_transient_for(self.exaile.window)
        self.action_area = self.xml.get_widget('track_editor_action_area')

        # FIXME: most of the metadata isn't stored in the database
        # so we have to load it explicitly from disk each time
        self.songs = []
        
        for song in tracks.get_selected_tracks():
            track = media.read_from_path(song.io_loc)
            for item in DB_COPY_ITEMS:
                if hasattr(song, item):
                    setattr(track, item, getattr(song, item))
            self.songs.append(track)
        
        self.count = len(self.songs)
        self.unsaved_tags = {}
        self.tags_to_add = {}
        self.selected_songs = editor.TrackGroup() 

        self.get_widgets()
        self.setup_left()
        self.setup_right()
        self.connect_events()

        self.tags_save.set_sensitive(False)
        self.tags_clear.set_sensitive(False)

        self.on_trackview_selection_changed(self.trackview)

        self.dialog.show()

    def get_widgets(self):
        """
            Gets all widgets from the glade definition file
        """
        xml = self.xml
        self.trackview = xml.get_widget('te_trackview')
        self.trackview_selection = self.trackview.get_selection()
        self.trackview_down = xml.get_widget('te_trackview_down')
        self.trackview_up = xml.get_widget('te_trackview_up')
        self.trackview_remove = xml.get_widget('te_trackview_remove')

        self.nb = xml.get_widget('te_nb')
        
        self.tag_view = xml.get_widget('te_tag_view')
        self.tags_add = xml.get_widget('te_tags_add')
        self.tags_remove = xml.get_widget('te_tags_remove')
        self.tags_clear = xml.get_widget('te_tags_clear')
        self.tags_save = xml.get_widget('te_tags_save')

        self.tagsfrompath_cb = xml.get_widget('te_path_patterns_cb')
        self.path_pattern_add = xml.get_widget('te_path_pattern_add')
        self.path_pattern_remove = xml.get_widget('te_path_pattern_remove')
        self.tagsfrompath_view = xml.get_widget('te_tagsfrompath_view')
        self.replace_tags1 = xml.get_widget('te_replace_tags1')
        self.replace_tags2 = xml.get_widget('te_replace_tags2')
        self.tagsfrompath_save = xml.get_widget('te_tagsfrompath_save')
        
        self.tracknum_start = xml.get_widget('te_tracknum_start')
        self.tracknum_total = xml.get_widget('te_tracknum_total')
        self.tracknum_view = xml.get_widget('te_tracknum_view')
        self.tracknum_save = xml.get_widget('te_tracknum_save')
        self.trackview = xml.get_widget('te_trackview')

    def connect_events(self):
        """
            Sets up the callbacks for all the events
        """
        self.trackview_down.connect('clicked', self.on_trackview_updown_clicked, False)
        self.trackview_up.connect('clicked', self.on_trackview_updown_clicked, True)
        self.trackview_remove.connect('clicked', self.on_trackview_remove_clicked)
        self.trackview_selection.connect('changed', self.on_trackview_selection_changed)

        self.nb.connect('switch_page', self.on_nb_page_switched)

        self.tags_clear.connect('clicked', self.on_tags_clear_clicked)
        self.tags_save.connect('clicked', self.on_tags_save_clicked)
        self.tags_remove.connect('clicked', self.on_tags_remove_clicked)
        self.tags_add.connect('clicked', self.on_tags_add_clicked)

        self.tagsfrompath_cb.connect('changed', self.on_path_patterns_changed)
        self.path_pattern_add.connect('clicked', self.on_path_pattern_add_clicked)
        self.path_pattern_remove.connect('clicked', self.on_path_pattern_remove_clicked)
        self.tagsfrompath_save.connect('clicked', self.on_tagsfrompath_save_clicked)

        self.tracknum_view.get_model().connect('row-deleted', \
            lambda *x: self.update_tracknum_view(True))

        self.tracknum_start.connect('output', \
            lambda *x: self.update_tracknum_sb())
        self.tracknum_start.connect('value-changed', \
            lambda *x: self.update_tracknum_view())

        self.tracknum_total.connect('output', \
            lambda *x: self.update_tracknum_sb())
        self.tracknum_total.connect('value-changed', \
            lambda *x: self.update_tracknum_view())

        self.tracknum_save.connect('clicked', self.on_tracknum_save_clicked)

    def setup_left(self):
        """
            Build the treeview in the left pane
        """
        tv = self.trackview
        liststore = gtk.ListStore(str, int)
        selection = self.trackview_selection
        selection.set_mode(gtk.SELECTION_MULTIPLE)
        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Tracks')

        tv.set_model(liststore)
        tv.append_column(col)
        col.pack_start(cell, True)
        col.add_attribute(cell, 'text', 0)
        tv.set_reorderable(True)

        for i, song in enumerate(self.songs):
            liststore.append([song.filename, i])

        selection.select_path((0,))
        return

    def setup_right(self):
        """
            Set up the treeviews in the tabs etc
        """
        # tags view
        # TRANSLATORS: A media file tag
        tag_column = gtk.TreeViewColumn(_("Tag"))
        tag_cell = gtk.CellRendererText()
        value_column = gtk.TreeViewColumn(_("Value"))
        value_cell = gtk.CellRendererText()
        value_cell.set_property("editable", True)
        value_cell.connect("edited", self.on_tag_cell_edited)

        # the tag itself, tag ID ('artist' instead of 
        # _('Artist')), the index (since every tag value is a
        # list), and whether a tag is
        # being removed, added or updated.
        model = gtk.ListStore(str, str, int, bool, bool, bool, bool)
        self.tag_view.set_model(model)

        self.tag_view.append_column(tag_column)
        tag_column.pack_start(tag_cell, True)
        tag_column.add_attribute(tag_cell, 'strikethrough', REMOVED)
        tag_column.set_cell_data_func(tag_cell, self.tagname_data_func)

        self.tag_view.append_column(value_column)
        value_column.pack_start(value_cell, True)
        value_column.add_attribute(value_cell, 'strikethrough', REMOVED)
        value_column.set_cell_data_func(value_cell, self.tagvalue_data_func)

        # track num view
        track_column = gtk.TreeViewColumn(_('Track'))
        track_cell = gtk.CellRendererText()
        num_column = gtk.TreeViewColumn(_('Track Number'))
        num_cell = gtk.CellRendererText()
        total_column = gtk.TreeViewColumn(_('Total Tracks'))
        total_cell = gtk.CellRendererText()

        # song object, title, tracknum, total tracks
        model = gtk.ListStore(object, str, int, int)
        self.tracknum_view.set_model(model)

        self.tracknum_view.append_column(track_column)
        track_column.pack_start(track_cell, True)
        track_column.add_attribute(track_cell, 'text', 1)

        self.tracknum_view.append_column(num_column)
        num_column.pack_start(num_cell, True)
        num_column.add_attribute(num_cell, 'text', 2)

        self.tracknum_view.append_column(total_column)
        total_column.pack_start(total_cell, True)
        total_column.add_attribute(total_cell, 'text', 3)

        # tags from path combobox and view

        def sep_func(model, iter, data=None):
            """
                Draw a separator after the default patterns
            """
            return int(model.get_string_from_iter(iter)) \
                        == len(TagsFromPattern.default_patterns)

        self.tagsfrompath_cb.set_row_separator_func(sep_func)
        self.populate_tagsfrompath_cb()
        self.tagsfrompath_cb.set_active(0)


        # this builds the model and everything, no need to do it here
        self.update_tagsfrompath_view()
        
        # update this view since its tab is displayed by default
        self.update_tag_view()

        return
    
    def tagname_data_func(self, column, cell, model, iter, *args):
        """
            Converts the tagname to a translated value
        """
        cell.set_property('markup', TAG_STRINGS[model.get_value(iter, TAGNAME)])
        
    def tagvalue_data_func(self, column, cell, model, iter, *args):
        """
            Sets all kinds of properties for the tag value
        """
        tag, value, oldvalue, error, added, removed, edited = \
            model[model.get_string_from_iter(iter)]

        if error:
            cell.set_property('markup', get_error(error))
        else:
            cell.set_property('markup', value)
        cell.set_property('editable', not removed and not error)

    def get_selected_tracks(self):
        """
            Returns a TrackGroup of the selected tracks
        """
        (liststore, paths) = self.trackview_selection.get_selected_rows()
        ret = []
        for path in paths:
            iter = liststore.get_iter(path)
            i = liststore.get_value(iter, 1)
            ret.append(self.songs[i])
        return editor.TrackGroup(ret)

    def update_tag_view(self):
        """
            Display tags
        """
        model = self.tag_view.get_model()
        self.tag_view.set_model(None)
        model = self.selected_songs.make_model()
        self.tag_view.set_model(model)
        self.tags_save.set_sensitive(False)
        self.tags_clear.set_sensitive(False)

    def update_tagsfrompath_view(self):
        """
            The most annoying treeview to update ;)
        """
        view = self.tagsfrompath_view
        if self.selected_songs: pattern_text = self.tagsfrompath_cb.get_active_text()
        else: 
            pattern_text = ""
        if not pattern_text: 
            self.tagsfrompath_view.set_model(None)
            return
        try: pattern = TagsFromPattern(pattern_text)
        except re.error: 
            xlmisc.log('Tags From Pattern: invalid pattern')
            return

        invalid = []
        for header in pattern.headers:
            if not min([track.can_change(header) \
                for track in self.selected_songs]):
                invalid.append(header)
        
        if len(invalid):
            msg = ngettext("All files currently selected do not support"
                           " editing the tag <b>%s</b>.",
                           "All files currently selected do not support"
                           " editing the tags <b>%s</b>.",
                           len(invalid))
            common.error(self.dialog, msg % ', '.join(invalid))
            pattern = TagsFromPattern('')

        view = self.tagsfrompath_view

        view.set_model(None)
        model = gtk.ListStore(object, str, *([str] * len(pattern.headers)))
        for col in view.get_columns():
            view.remove_column(col)
        
        col = gtk.TreeViewColumn(_('File'), gtk.CellRendererText(), text=1)
        col.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        view.append_column(col)

        for i, header in enumerate(pattern.headers):
            render = gtk.CellRendererText()
            render.set_property('editable', True)
            render.connect('edited', self.tagsfrompath_row_edited, model, i + 2)
            col = gtk.TreeViewColumn(TAG_STRINGS[header], render, text=i + 2)
            col.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
            view.append_column(col)

        for song in self.selected_songs:
            basename = os.path.basename(song.io_loc)
            row = [song, basename]
            match = pattern.match(song)
            for h in pattern.headers:
                text = match.get(h, '')
                #for f in self.filters:
                #    if f.active: text = f.filter(h, text)
                if not song.is_multi():
                    text = u", ".join(text.split("\n"))
                row.append(text)
            model.append(row=row)

        view.set_model(model)
        return

    def tagsfrompath_row_edited(self, renderer, path, new, model, colnum):
        row = model[path]
        if row[colnum] != new:
            row[colnum] = new

    def update_tracknum_view(self, reordered=False):
        """
            Update the track numbering view. If reordered is True,
            it means we will preserve the order of the songs.
        """
        self.update_tracknum_sb()
        start = self.tracknum_start.get_value()
        total = self.tracknum_total.get_value()
        #if total == 0: total = None
        model = self.tracknum_view.get_model()

        if reordered:
            for i, row in enumerate(model):
                row[2] = i + start
        else:
            model.clear()
            for i, song in enumerate(self.selected_songs):
                model.append([song, song.title, i + start, total])
        
        return

    def populate_tagsfrompath_cb(self):
        """
            Constructs the combobox in the Tags from Path tab
        """
        cb = self.tagsfrompath_cb
        cb.set_active(-1)
        cb.get_model().clear()
        map(cb.append_text, TagsFromPattern.default_patterns)
        patterns = self.exaile.settings.get_list('tagsfrompath_patterns', [])

        if patterns:
            map(cb.append_text, [''] + patterns)
        else:
            return

    def on_nb_page_switched(self, nb, page, page_num, *args):
        """
            Update the affected tree view when the page is changed
        """
        if self.check_changed():
            msg = _('Do you want to save your changes in the Tags tab?')
            if common.yes_no_dialog(self.dialog, msg) == gtk.RESPONSE_YES:
                self.write_tags()
                self.update_tag_view()
        return

    def on_trackview_remove_clicked(self, widget, *args):
        """
            Remove the selected tracks from the list
        """
        selection = self.trackview_selection
        (liststore, paths) = selection.get_selected_rows()
        
        paths.sort(reverse=True)
        for path in paths:
            iter = liststore.get_iter(path)
            self.songs.pop(liststore.get_value(iter, 1))
            liststore.remove(iter)
        selection.select_path(0)

    def on_trackview_updown_clicked(self, widget, up):
        """
            Move the selection one step up or down
        """
        selection = self.trackview_selection
        (liststore, paths) = selection.get_selected_rows()
        size = len(self.trackview)

        if not paths: return

        paths.sort()
        old = int(paths[0][0])

        if (up and old == 0) or (not up and old == size - 1):
            return
        else:
            if up:
                new = old - 1
            else:
                new = old + 1

        selection.unselect_all()
        selection.select_path(str(new))

    def check_changed(self):
        """
            Check whether anything has changed in the tag view
        """
        model = self.tag_view.get_model()
        for row in model:
            if True in [row[ADDED], row[REMOVED], row[EDITED]]:
                return True
        return False

    def on_trackview_selection_changed(self, widget, *args):
        """
            Update the track list inside the active tab
        """
        if self.check_changed():
            msg = _('Do you want to save your changes in the Tags tab?')
            if common.yes_no_dialog(self.dialog, msg) == gtk.RESPONSE_YES:
                self.write_tags()

        self.selected_songs = self.get_selected_tracks()

        self.update_tag_view()
        self.update_tagsfrompath_view()
        self.update_tracknum_view()
        return

    def on_tag_cell_edited(self, widget, path, new_text, *args):
        """
            Gets called whenever the user has edited a row in the tag view.
        """
        if new_text == widget.get_property('text'): return
        
        model = self.tag_view.get_model()
        iter = model.get_iter(path)

        oldval = model.get_value(iter, VALUE)
        model.set_value(iter, OLDVALUE, oldval)
        model.set_value(iter, VALUE, new_text)
        model.set_value(iter, EDITED, True)

        self.tags_save.set_sensitive(True)
        self.tags_clear.set_sensitive(True)

    def on_tags_clear_clicked(self, widget, *args):
        """
            Remove the changes from the tag view
        """
        self.update_tag_view()

    def on_tags_save_clicked(self, widget, *args):
        """
            Write changes and refresh the affected rows in the main window
        """
        self.write_tags()
        self.update_tag_view()

    def write_tags(self):
        """
            Write the changes to disk
        """
        model = self.tag_view.get_model()
        errors = self.selected_songs.write_tags(self.exaile, model)

        if errors:
            message = ""
            count = 1
            for error in errors:
                message += "%d: %s\n" % (count, error)
                count += 1
            self.dialog.hide()
            common.scrolledMessageDialog(self.exaile.window, message, _("Some errors"
                " occurred")) 

    def on_tags_add_clicked(self, widget, *args):
        """
            Add a tag to the track(s)
        """
        dialog = AddTagDialog(self.dialog)
        
        if dialog.run() != gtk.RESPONSE_OK:
            dialog.destroy()
            return
        try:
            tag = dialog.get_tag()
            value = dialog.get_value()

            if self.selected_songs.get_tag_status(tag) \
                == editor.TAG_UNSUPPORTED:
                raise xlmisc.TagUnsupportedException(tag)

            row = [tag, value, None, editor.TAG_NO_ERROR, True, False, False]
            self.tag_view.get_model().append(row)
            self.tags_save.set_sensitive(True)
            self.tags_clear.set_sensitive(True)
        except xlmisc.TagUnsupportedException, e:
            msg =_("The tag <b>%s</b> is not supported"
                   " by all selected tracks") % e.tag
            common.error(self.dialog, msg)
        except:
            xlmisc.log_exception()
        dialog.destroy()

        return

    def on_tags_remove_clicked(self, widget, *args):
        """
            Set the tag in question to None and remove from the tag list
        """
        (model, iter) = self.tag_view.get_selection().get_selected()
        model.set_value(iter, REMOVED, True)
        self.tags_save.set_sensitive(True)
        self.tags_clear.set_sensitive(True)
        
    def on_path_patterns_changed(self, widget, *args):
        """
            Update the treeview
        """
        self.update_tagsfrompath_view()
        return

    def on_path_pattern_add_clicked(self, widget, *args):
        """
            Pop up a dialog so the user can add a new path pattern
        """
        dialog = gtk.Dialog()
        dialog.add_buttons(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, \
            gtk.STOCK_OK, gtk.RESPONSE_OK)
        entry = gtk.Entry()
        # Set the default text to the pattern currently selected
        entry.set_text(self.tagsfrompath_cb.get_active_text())
        dialog.vbox.pack_start(entry, True, True)
        
        dialog.show_all()
        resp = dialog.run()
        val = unicode(entry.get_text(), 'utf-8')
        dialog.destroy()
        if resp != gtk.RESPONSE_OK or not val:
            return

        patterns = self.exaile.settings.get_list('tagsfrompath_patterns', [])
        patterns.append(val)
        self.exaile.settings.set_list('tagsfrompath_patterns', patterns)
        self.populate_tagsfrompath_cb()
        act = len(patterns + TagsFromPattern.default_patterns) 
        self.tagsfrompath_cb.set_active(act) # select the last item

    def on_path_pattern_remove_clicked(self, widget, *args):
        """
            Remove the pattern from the combobox, if it's not a default one.
        """
        active = self.tagsfrompath_cb.get_active()
        l = len(TagsFromPattern.default_patterns)
        if active > l:
            patterns = self.exaile.settings.get_list('tagsfrompath_patterns', [])
            del(patterns[active - l - 1])
            self.exaile.settings.set_list('tagsfrompath_patterns', patterns)
            self.populate_tagsfrompath_cb()
            self.tagsfrompath_cb.set_active(0)
        elif active < l:
            common.error(self.dialog, _("The default patterns can't be removed."))
            pass

    def on_tagsfrompath_save_clicked(self, widget, *args):
        """
            Write to disk. If a value is None, don't write it
        """
        # FIXME: how do we know when a click on 'save' will blank
        # out a lot of valuable tags and when it's a wanted effect?
        append = self.replace_tags2.get_active()
        cols = self.tagsfrompath_view.get_columns()
        errors = []
        exaile = self.exaile
        for row in self.tagsfrompath_view.get_model():
            track = row[0]
            for i, col in enumerate(cols[1:]):
                if not row[i+2]: continue
                tag = (col.get_title()).lower().replace(' ', '')
                track.set_tag(tag, row[i + 2], append)
            try:
                media.write_tag(track)
                library.save_track_to_db(exaile.db, track)
                try:
                    exaile.all_songs.remove(exaile.all_songs.for_path(track.loc))
                    exaile.all_songs.append(track)
                except:
                    xlmisc.log_exception()

                exaile.tracks.refresh_row(track)
            except:
                errors.append(_("Unknown error writing tag for %s") % track.loc)
                xlmisc.log_exception()
        return

    def on_tracknum_save_clicked(self, widget, *args):
        """
            Write to disk
        """
        errors = []
        exaile = self.exaile
        
        for row in self.tracknum_view.get_model():
            track = row[0]
            if row[3]:
                track.set_tag('tracknumber', "/".join([str(row[2]), str(row[3])]))
            else:
                track.set_tag('tracknumber', row[2])
            try:
                media.write_tag(track)
                library.save_track_to_db(exaile.db, track)
                try:
                    exaile.all_songs.remove(exaile.all_songs.for_path(track.loc))
                    exaile.all_songs.append(track)
                except:
                    xlmisc.log_exception()

                exaile.tracks.refresh_row(track)
            except:
                errors.append(_("Unknown error writing tag for %s") % track.loc)
                xlmisc.log_exception()
        return

    def update_tracknum_sb(self):
        """
            Make sure we don't get any weird values for our track numbering
        """
        start = self.tracknum_start.get_value()
        total = self.tracknum_total.get_value()

        l = len(self.selected_songs)
        if total < (start + l - 1): 
            self.tracknum_total.set_value(start + l - 1)

class TagsFromPattern(object):
    """
        Gratefully borrowed from Quod Libet
        FIXME: you can't have the same tag twice (hard to fix)
    """
    default_patterns = ['<artist>/<album>/<tracknumber> - <title>', 
                        '<tracknumber>. <title>',
                        '<tracknumber> - <title>',
                        '<tracknumber> - <artist> - <title>',
                        '<artist> - <album>/<tracknumber>. <title>']
    def __init__(self, pattern):
        self.compile(pattern)

    def compile(self, pattern):
        self.headers = []
        self.slashes = len(pattern) - len(pattern.replace(os.path.sep,'')) + 1
        self.pattern = None

        # patterns look like <tagname> non regexy stuff <tagname> ...
        pieces = re.split(r'(<[A-Za-z0-9_]+>)', pattern)
        override = { '<tracknumber>': r'\d\d?', '<discnumber>': r'\d\d??' }
        for i, piece in enumerate(pieces):
            if not piece: continue
            if piece[0]+piece[-1] == '<>' and piece[1:-1].isalnum():
                piece = piece.lower()   # canonicalize to lowercase tag names
                pieces[i] = '(?P%s%s)' % (piece, override.get(piece, '.+?'))
                self.headers.append(piece[1:-1].encode("ascii", "replace"))
            else:
                pieces[i] = re.escape(piece)
                
        # some slight magic to anchor searches "nicely"
        # nicely means if it starts with a <tag>, anchor with a /
        # if it ends with a <tag>, anchor with .xxx$
        # but if it's a <tagnumber>, don't bother as \d+ is sufficient
        # and if it's not a tag, trust the user
        if pattern.startswith('<') and not pattern.startswith('<tracknumber>')\
                and not pattern.startswith('<discnumber>'):
            pieces.insert(0, os.path.sep)
        if pattern.endswith('>') and not pattern.endswith('<tracknumber>')\
                and not pattern.endswith('<discnumber>'):
            pieces.append(r'(?:\.[A-Za-z0-9_+]+)$')

        self.pattern = re.compile(''.join(pieces))

    def match(self, song):
        song = song.io_loc
        # only match on the last n pieces of a filename, dictated by pattern
        # this means no pattern may effectively cross a /, despite .* doing so
        sep = os.path.sep
        matchon = sep+sep.join(song.split(sep)[-self.slashes:])
        match = self.pattern.search(matchon)

        # dicts for all!
        if match is None: return {}
        else: return match.groupdict()
                
class AddTagDialog(gtk.Dialog):
    """
        Dialog shown when the user wants to add a tag
    """
    def __init__(self, parent):
        super(AddTagDialog, self).__init__(_("Add a Tag"), parent)
        self.set_border_width(3)
        self.set_has_separator(False)
        self.set_resizable(False)
        self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        self.add = self.add_button(gtk.STOCK_ADD, gtk.RESPONSE_OK)
        self.add.set_sensitive(False)
        self.set_default_response(gtk.RESPONSE_OK)

        self.combo = gtk.combo_box_new_text()
        self.populate_combo()
        self.combo.set_active(0)
        self.entry = gtk.Entry()
        self.entry.connect('changed', self.on_entry_changed)

        self.vbox.pack_start(self.combo, True, True)
        self.vbox.pack_start(self.entry, True, True)
        self.show_all()

    def populate_combo(self):
        """
            Place the tags in the combobox
        """
        for tag in TAG_STRINGS.values():
            self.combo.append_text(tag)

    def on_entry_changed(self, entry, *args):
        """
            Check if there's any text in the entry, set add button sensitivity
        """
        if entry.get_property('text'):
            self.add.set_sensitive(True)
        else:
            self.add.set_sensitive(False)

    def get_tag(self):
        """
            Returns the tag that the user has selected
        """
        return TAG_STRINGS.keys()[self.combo.get_active()]

    def get_value(self):
        """
            Returns the value that the user has inserted
        """
        return self.entry.get_property('text')

    def run(self):
        self.show()
        return super(AddTagDialog, self).run()
        
class AskSaveDialog(gtk.Dialog):
    """
        Ask the user if he wants to save his changes or not
    """
    def __init__(self, parent):
        gtk.Dialog.__init__(self, _("Save changes?"), parent)
        return

def update_rating(caller, num, tracks=None):
    """
        Updates the rating based on which menu id was clicked
    """
    rating = num + 1

    cur = caller.db.cursor()
    if tracks is None: tracks = caller.get_selected_tracks()
    for track in tracks:
        path_id = library.get_column_id(caller.db, 'paths', 'name',
            track.loc)
        caller.db.execute("UPDATE tracks SET user_rating=? WHERE path=?",
            (rating, path_id)) 
        track.rating = rating
        if isinstance(caller, trackslist.TracksListCtrl):
            caller.refresh_row(track)

def get_error(i):
    """
        Returns a textual error with pango tags for the tag view
    """
    before = '<span foreground="gray" style="italic">'
    after = '</span>'
    if i == editor.TAG_MISSING: 
        ret = _("Tag missing from one or more files")
    elif i == editor.TAG_DIFFERENT: 
        ret = _("Tag different in one or more files")
    elif i == editor.TAG_UNSUPPORTED: 
        ret = _("Tag unsupported in one or more files")
    elif i == editor.TAG_MULTI: 
        ret = _("Multiple tag values found, not supported by all tracks")
    else:
        ret = _("Unable to display tag for unknown reason")

    return before + ret + after



#!/usr/bin/python

import sys
import time
import socket
import md5
import threading
import cgi
import os
import string

import config
import lastfm
from lib import scrobbler

True = 1
False = 0


class proxy:

    def __init__(self, username, password):
        self.username = username
        self.password = password
        self.lastfm = None
        self.audioscrobbler = None
        self.bookmarks = []
        self.lasttracks = []
        self.basedir = "."
        self.quit = False
        self.skip = 0
        self.stop = False
        self.version = "1.3b"
        self.streaming = 0
        self.recordtoprofile = 1
        self.starttime = 0
        self.rating = ""
        self.proxy_ready = False
        self.np_image_func = None
        self.last_cover = ''

        self.extraheaders = string.join([
                "Server: LastFMProxy/" + self.version + "\r\n",
                "Pragma: no-cache\r\n",
                "Cache-Control: max-cache=0\r\n"
                ], "");

    def hexify(self, s):
        result = ""
        for c in s:
            result = result + ("%02x" % ord(c))
        return result

    def storebookmark(self, station):
        newbookmarks = []
        newbookmarks.append(station)
        for s in self.bookmarks:
            if s != station:
                newbookmarks.append(s)

        self.bookmarks = newbookmarks
        if len(self.bookmarks) > 10:
            self.bookmarks = self.bookmarks[:10]

        f = open(os.path.join(self.basedir, "bookmarks.txt"), "w")
        for s in self.bookmarks:
            f.write(s + "\n")
        f.close()

    def storetrack(self, track):
        if len(self.lasttracks) > 0 and track == self.lasttracks[0]:
            return
        self.lasttracks = ([ track ] + self.lasttracks)[:6]

    def gotconnection(self, clientsock):
        metaint = 1024*32
        sendmetadata = False

        req = ""
        while True:
            data = clientsock.recv(1)
            if data == "":
                break
            req = req + data
            if req[-4:] == "\r\n\r\n":
                break
        req = string.split(req, "\n")
        http = {}
        for line in req:
            tmp = string.split(line, ":", 1)
            #print tmp
            if len(tmp) == 2:
                http[tmp[0]] = string.strip(tmp[1])
                if string.lower(tmp[0]) == "icy-metadata":
                    sendmetadata = int(tmp[1])

        # Make sure http "Host" variable is fully qualified
        if not http.has_key("Host"):
            http["Host"] = "localhost:" + str(config.listenport)
        elif string.find(http["Host"], ":") < 0:
            http["Host"] = http["Host"] + ":" + str(config.listenport)

        # Check request method
        req[0] = string.split(req[0], " ")
        if len(req[0]) != 3 or req[0][0] != "GET":
            print "Unhandled method:", req[0]
            try:
                clientsock.sendall("HTTP/1.0 405 Method Not Allowed\r\n");
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Allow: GET\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall("Unhandled method.\r\n")
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        tmp = string.split(req[0][1], "?", 1)
        station = tmp[0]
        if len(tmp) > 1:
            args = cgi.parse_qs(tmp[1])
        else:
            args = None

        if self.lastfm.debug:
            sys.stderr.write("station=" + station + "\n")

        if station == "/frames":
            cont = "<frameset rows=\"220, *\"><FRAME src=\"/\"><frame src=\"http://www.last.fm/\"></frameset>\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/quit":
            self.quit = True
            cont = "Bye!\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station[:9] == "/settheme":
            theme = station[10:]
            try:
                tmp = os.stat(os.path.join(self.basedir, "data", theme + ".css"))
                tmp = os.stat(os.path.join(self.basedir, "data", theme + ".html"))
                config.theme = theme
                url = "http://" + http["Host"] + "/"
            except OSError:
                url = "http://" + http["Host"] + "/?msg=Theme%20" + theme + "%20not%20found!"
                
            cont = "Moved to <a href=\"" + url + "\">here</a>."

            try:
                clientsock.sendall("HTTP/1.0 307 Temporary Redirect\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Location: " + url + "\r\n")
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station[-4:] == ".pls":
            cont = "[playlist]\n"
            cont = cont + "File1=http://" + http["Host"] + string.replace(station, ".pls", ".mp3") + "\n"
            cont = cont + "Title1=Last.FM Radio\n"
            cont = cont + "Length1=-1\n"
            cont = cont + "NumberOfEntries=1\n"
            cont = cont + "Version=2\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: audio/x-scpls\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station[-4:] == ".m3u":
            cont = "http://" + http["Host"] + string.replace(station, ".m3u", ".mp3") + "\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: audio/x-mpegurl\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/skip" or station == "/love" or station == "/ban" or station[:8] == "/lastfm:" or station[:22] == "/changestation/lastfm:" or station == "/rtp" or station == "/nortp" or station == "/skip":

            redirect = 0

            if station == "/rtp":
                self.recordtoprofile = 1
                res = {"response": "OK, scrobbling enabled."}

            elif station == "/nortp":
                self.recordtoprofile = 0
                res = {"response": "OK, scrobbling disabled."}

            elif station == "/skip":
                self.skip = 1
                self.rating = "S"
                res = {"response": "OK, skipping..."}

            elif station[:8] == "/lastfm:":
                res = self.lastfm.changestation(station[1:])
                if res.has_key("response") and res["response"] == "OK":
                    self.skip = 1000
                    self.storebookmark(station[1:])
                redirect = 1

            # /changestation is used by javascript and should not be redirected
            elif station[:22] == "/changestation/lastfm:":
                res = self.lastfm.changestation(station[15:])
                if res.has_key("response") and res["response"] == "OK":
                    res["response"] = "OK, changing station..."
                    self.skip = 1000
                    self.storebookmark(station[15:])

            else:
                if station == "/love":
                    self.rating = "L"
                elif station == "/ban":
                    self.rating = "B"
                    self.skip = 1

                res = self.lastfm.command(station[1:])

            if res.has_key("response"):
                res = res["response"]
            else:
                print "hmm?", repr(res)
                res = "INTERNALERROR"

            cont = "result = '" + res + "';\n"

            try:
                if redirect:
                    clientsock.sendall("HTTP/1.0 307 Temporary Redirect\r\n")
                    clientsock.sendall("Location: http://" + http["Host"] + "/\r\n")
                else:
                    clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/plain\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/popup":
            cont = "<script>\nfunction popup(url,w,h) {\nwindow.open(url, \"LastFMProxy\", 'width='+w+',height='+h+'toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=no,resizable=yes');\n}\n</script>\n"
            cont = cont + "<div><a href=\"javascript:popup('/',700,300);\">Click here to open popup-window</a></div>\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/np":

            cont = u"np_streaming = " + str(self.streaming) + ";\n"

            if self.streaming:
               
                try:
                    title = self.lastfm.playlist.data.title
                except:
                    title = "Wait..."

                cont = cont + "np_station = '" + title + "';\n"
                
                try:
                    m = self.lastfm.playlist.data.tracks[self.lastfm.playlist.pos]
                except:
                    m = False

                if m and m.has_key("trackpage") and m.has_key("creator") and m.has_key("title"):
                    tmp = "<a href=\"" + m["trackpage"] + "\">"
                    tmp = tmp + m["creator"].decode("utf8") + " - "
                    tmp = tmp + m["title"].decode("utf8") + "</a>"
                    self.storetrack(tmp)

                if m:
                    for tmp in m.keys():
                        # LASTFMPROXY:  Set album cover
                        if tmp == 'image':
                            if self.last_cover != m[tmp]:
                                self.np_image_func(m[tmp].decode('utf8'))
                            self.last_cover = m[tmp]
                        cont = cont + "np_" + tmp + " = '" + string.replace(m[tmp].decode("utf8"), "'", "\\'") + "';\n"

                cont = cont + "np_trackprogress = " + str(int(time.time() - self.starttime)) + ";\n"

            cont = cont + "np_recordtoprofile = " + str(self.recordtoprofile) + ";\n"

            cont = cont + "np_lasttracks = " + str(len(self.lasttracks)) + ";\n"
            cont = cont + "np_lasttrack = new Array(" + str(len(self.lasttracks)) + ");\n"
            for i in range(0,len(self.lasttracks)):
                cont = cont + "np_lasttrack[" + str(i) + "] = '" + string.replace(self.lasttracks[i], "'", "\\'") + "';\n"

            cont = cont + "np_bookmarks = " + str(len(self.bookmarks)) + ";\n"
            cont = cont + "np_bookmark = new Array(" + str(len(self.bookmarks)) + ");\n"
            for i in range(0,len(self.bookmarks)):
                cont = cont + "np_bookmark[" + str(i) + "] = '" + string.replace(self.bookmarks[i], "'", "\\'") + "';\n"

            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/plain\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont.encode("utf8"))
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/debug" or station == "/nodebug":
            self.lastfm.debug = station == "/debug"
            cont = "debug=" + str(self.lastfm.debug) + "\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/info":
            # Dump current metadata struct
            cont = "<p><pre>"
            for tmp in self.lastfm.info.keys():
                cont = cont + tmp + ": " + self.lastfm.info[tmp] + "\n"
            cont = cont + "</pre></p>\n"
            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/html\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station == "/":
            title = "LastFMProxy v" + self.version
            cont = "<html><head><title>" + title + "</title>\n"
            cont = cont + "<link rel=\"shortcut icon\" href=\"/data/favicon.ico\" />\n"
            cont = cont + "<link rel=\"icon\" href=\"/data/favicon.ico\" />\n"
            cont = cont + "<link rel=\"icon\" type=\"image/png\" href=\"/data/nice_favicon.png\" />\n"
            cont = cont + "<link rel=\"stylesheet\" type=\"text/css\" media=\"screen\" href=\"data/" + config.theme + ".css\" />\n"
            cont = cont + "<script>\n"
            cont = cont + "var host = 'http://" + http["Host"] + "';\n"
            f = open(os.path.join(self.basedir, "data", "main.js"), "r")
            cont = cont + f.read()
            f.close()
            cont = cont + "</script>\n"

            cont = cont + "</head><body>\n"
            cont = cont + "<form action=\"/\" name=\"lfmpform\" method=\"get\">"

            f = open(os.path.join(self.basedir, "data", config.theme + ".html"), "r")
            gui = f.read()
            f.close()

            tmp = "<a href=\"http://vidar.gimp.org/lastfmproxy/\" target=\"_blank\">LastFMProxy v" + self.version + "</a> - &copy; 2005-2007 Vidar Madsen"
            gui = string.replace(gui, "VERSION", tmp)

            gui = string.replace(gui, "STATION", "<span id=\"lfmp-dyn-station\">&nbsp;</span>")

            tmp = "<span id=\"lfmp-dyn-lasttracks-list\"></span>"
            gui = string.replace(gui, "LASTTRACKS", tmp)

            cover = "data/noalbum_medium.gif"
            gui = string.replace(gui, "COVER", "<span id=\"lfmp-dyn-cover\"><img width=130 height=130 src=\"" + cover + "\"></span>")
            gui = string.replace(gui, "SIMILAR", "<span id=\"lfmp-dyn-similarlink\">&nbsp;</span>")
            gui = string.replace(gui, "FANS", "<span id=\"lfmp-dyn-fanslink\">&nbsp;</span>")
            gui = string.replace(gui, "ARTIST", "<span id=\"lfmp-dyn-artist\">&nbsp;</span>")
            gui = string.replace(gui, "ALBUM", "<span id=\"lfmp-dyn-album\">&nbsp;</span>")
            gui = string.replace(gui, "TRACK", "<span id=\"lfmp-dyn-track\">&nbsp;</span>")
            gui = string.replace(gui, "DURATION", "<span id=\"lfmp-dyn-dur\">&nbsp;</span>")

            tmp = "<select name=\"stationselect\" onChange=\"selectstation();\"><option value=\"\">...</option></select>"
            gui = string.replace(gui, "BOOKMARKS", tmp)

            tmp = "<a href=\"/\" target=\"_self\" title=\"Refresh (R)\">Refresh</a> &middot; "

            tmp = tmp + "<span id=\"lfmp-buttons1\" style=\"display: none;\">"
            tmp = tmp + "<a href=\"javascript:skip();\" title=\"Skip track (space)\">Skip</a> &middot; "
            tmp = tmp + "<a href=\"javascript:love();\" title=\"Love this track (enter)\">Love</a> &middot; "
            tmp = tmp + "<a href=\"javascript:ban();\" title=\"Ban this track (backspace)\">Ban</a> &middot; "
            tmp = tmp + "<input type=\"checkbox\" onChange=\"togglertp();\" name=\"rtp\" id=\"lfmp-rtp\"> <label for=\"lfmp-rtp\">R<span class=\"lfmp-shortcut\">e</span>cord to profile</label> &middot; "
            tmp = tmp + "</span>"

            tmp = tmp + "<span id=\"lfmp-buttons2\" style=\"display: none;\">"
            tmp = tmp + "<a href=\"http://" + http["Host"] + "/lastfm.m3u\">Start radio</a>\n"
            tmp = tmp + "</span>"

            gui = string.replace(gui, "BUTTONS", tmp)

            cont = cont + gui

            cont = cont + "</form>"
            cont = cont + "<script>\ntick();\n</script>\n"
            cont = cont + "</body></html>\n"

            try:
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/html; charset=utf-8\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station[:6] == "/data/":
            if station[-4:] == ".css":
                type = "text/css"
            elif station[-4:] == ".png":
                type = "image/png"
            elif station[-4:] == ".jpg":
                type = "image/jpeg"
            elif station[-4:] == ".gif":
                type = "image/gif"
            else:
                type = "application/octet-stream"

            try:
                f = open(os.path.join(self.basedir, "data", station[6:]), "rb")
            except IOError:
                cont = "File \"" + station + "\" not found\n"
                try:
                    clientsock.sendall("HTTP/1.0 404 Not Found\r\n")
                    clientsock.sendall(self.extraheaders)
                    clientsock.sendall("Content-Type: text/plain\r\n")
                    clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                    clientsock.sendall("\r\n")
                    clientsock.sendall(cont)
                    clientsock.close()
                except socket.error:
                    clientsock.close()
                return

            cont = f.read()
            f.close()
            try:
                # No extraheaders here, since we want to cache these files.
                clientsock.sendall("HTTP/1.0 200 OK\r\n")
                clientsock.sendall("Server: LastFMProxy/" + self.version + "\r\n")
                clientsock.sendall("Content-Type: " + type + "\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return

        if station[-4:] != ".mp3":
            cont = "File \"" + station + "\" not found\n"
            try:
                clientsock.sendall("HTTP/1.0 404 Not Found\r\n")
                clientsock.sendall(self.extraheaders)
                clientsock.sendall("Content-Type: text/plain\r\n")
                clientsock.sendall("Content-Length: " + str(len(cont)) + "\r\n")
                clientsock.sendall("\r\n")
                clientsock.sendall(cont)
                clientsock.close()
            except socket.error:
                clientsock.close()
            return
        
        self.skip = 0
        self.stop = False
        retries = 0

        # Check if station is "last.mp3", and if not, send a station change
        if station != "/lastfm.mp3":
            station = "lastfm:/" + string.replace(station, ".mp3", "")
            #print "station:", station
            res = self.lastfm.changestation(station)
            if res.has_key("response") and res["response"] == "OK":
                self.storebookmark(station)
                self.skip = 1000
            #print repr(res)


        self.streaming = 1

        # Send HTTP headers to client
        try:
            clientsock.sendall("HTTP/1.1 200 OK\r\n");
            clientsock.sendall("Content-Type: audio/mpeg\r\n")
            clientsock.sendall("Cache-Control: no-cache, must-revalidate\r\n");
            if sendmetadata:
                for tmp in [ "icy-name:last.fm", 
                        "icy-pub:1", 
                        "icy-url:http://www.last.fm/",
                        "icy-metaint:" + str(metaint) ]:
                    clientsock.sendall(tmp + "\r\n")
            clientsock.sendall("\r\n")
        except:
            print "Error sending HTTP headers"
            self.stop = True

        while not self.quit and not self.stop:
                
            playlist = self.lastfm.playlist

            if self.skip or not playlist.data:
                playlist.pos = playlist.pos + self.skip
                if not playlist.data or playlist.pos >= len(playlist.data.tracks):
                    if not self.lastfm.getplaylist():
                        
                        print "Strangeness! No tracks in playlist? Sending station update..."
                        retries = retries + 1
                        if retries < 3:
                            if len(self.bookmarks) >= retries:
                                tmpstation = self.bookmarks[retries-1]
                            else:
                                tmpstation = "lastfm://user/" + self.username + "/neighbours"

                            print "Trying station " + tmpstation
                            res = self.lastfm.changestation(tmpstation)
                            if res.has_key("response") and res["response"] == "OK":
                                if self.lastfm.getplaylist():
                                    retries = 0
                                    continue

                        print "Unable to change station or get a playlist. Stopping."
                        self.stop = 1
                        break

                    playlist = self.lastfm.playlist
                self.skip = 0

            track = playlist.data.tracks[playlist.pos]

            url = track["location"]
            url = "%s" % url
            url = string.split(url, "/", 3)

            host = string.split(url[2], ":")
            if len(host) != 2:
                host = [ host[0], 80 ]
            else:
                host[1] = int(host[1])

            if self.lastfm.debug:
                sys.stderr.write("GET http://" + host[0] + ":" + str(host[1]) + "/" + url[3] + " HTTP/1.0\r\n")

            # Connect to actual server and request stream
            streamsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if config.useproxy:
                streamsock.connect((config.proxyhost, config.proxyport))
                streamsock.sendall("GET http://" + host[0] + ":" + str(host[1]) + "/" + url[3] + " HTTP/1.0\r\n")
            else:
                streamsock.connect((host[0], host[1]))
                streamsock.sendall("GET /" + url[3] + " HTTP/1.0\r\n")
            streamsock.sendall("Host: " + host[0] + "\r\n")
            streamsock.sendall("\r\n")

            # Read HTTP headers
            while True:
                line = ""
                while True:
                    c = streamsock.recv(1)
                    line = line + c
                    if c == '\n':
                        break

                if self.lastfm.debug:
                    sys.stderr.write("<<< " + line)

                # Handle "403 Invalid ticket" more gracefully
                if line[:6] == "HTTP/1":
                    tmp = string.split(line, " ", 3)
                    if tmp[1] != "200":
                        self.skip = 1
                        break

                if line == "\r\n":
                    break

            if self.skip:
                continue

            self.starttime = int(time.time())
            self.rating = ""

            tracklen = int(int(track["duration"])/1000)
            if self.recordtoprofile:
                self.audioscrobbler.now_playing(track["creator"], track["title"], 
                    track["album"], tracklen)

            scrobbled = False

            if self.lastfm.debug:
                sys.stderr.write("Starting stream...\n")

            # Stream data
            count = 0
            while not self.quit and not self.stop:
                l = metaint - count

                if self.skip:
                    data = ""
                else:
                    try:
                        data = streamsock.recv(l)
                    except socket.error:
                        print "Error receiving data from server"
                        data = ""

                if data == "":

                    # Send padding data so that metadata is offset correctly
                    if sendmetadata:
                        try:
                            clientsock.sendall("\0" * l)
                        except socket.error, (val,msg):
                            print "Error sending data to client:", msg
                            self.stop = 1
                            break

                        icytag = "StreamTitle='';StreamUrl='';" + chr(0) * 16
                        blocks = len(icytag) / 16
                        icytag = chr(blocks) + icytag[:blocks*16]
                        try:
                            clientsock.sendall(icytag)
                        except socket.error, (val,msg):
                            print "Error sending data to client:", msg
                            self.stop = 1
                            break

                    if not self.skip:
                        self.skip = 1
                    break

                #if self.lastfm.debug:
                #    sys.stderr.write("Got %d bytes\n" % len(data))

                count = count + len(data)
                if count == metaint:
                    count = 0

                try:
                    clientsock.sendall(data)
                except socket.error, (val,msg):
                    print "Error sending data to client:", msg
                    self.stop = 1
                    break

                # Inject Shoutcast tags
                if sendmetadata and count == 0:

                    trackname = track["creator"] + " - " + track["title"]

                    icytag = "StreamTitle='" + trackname + "';StreamUrl='';" + chr(0) * 16

                    blocks = len(icytag) / 16
                    icytag = chr(blocks) + icytag[:blocks*16]
                    try:
                        clientsock.sendall(icytag)
                    except socket.error, (val,msg):
                        print "Error sending data to client:", msg
                        self.stop = 1
                        break


            if self.lastfm.debug:
                sys.stderr.write("Stopped.\n")

            streamsock.close()

            if not scrobbled and self.recordtoprofile and not self.stop and not self.quit:
                if tracklen > 30 and (time.time() - self.starttime) > tracklen / 2:
                    scrobbled = True
                    self.audioscrobbler.submit(artist=track["creator"], 
                        track=track["title"], album=track["album"], time=self.starttime, 
                        rating=self.rating, length=tracklen, source='L' + track['trackauth'], 
                        autoflush=True)

        clientsock.close()
        self.streaming = 0

    def run(self, bind_address, port):
        print "Starting LastFMProxy " + self.version + "..."

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Set socket options to allow binding to the same address
        # when the proxy has been closed and then restarted:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((bind_address, port))
        s.listen(5)

        print "Connecting to last.fm server..."
        self.lastfm = lastfm.lastfm();
        self.lastfm.connect(self.username, self.hexify(md5.md5(self.password).digest()))
        #print self.lastfm.info

        if not self.lastfm.info.has_key("session"):
            print "Handshake failed."
            print "DEBUG:", self.lastfm.info
            s.close()
            return

        if self.lastfm.info["session"] == "FAILED":
            print "Handshake failed. Bad login info, perhaps?"
            print "DEBUG:", self.lastfm.info
            s.close()
            return

        self.audioscrobbler = scrobbler

        self.bookmarks = []
        try:
            f = open(os.path.join(self.basedir, "bookmarks.txt"), "r")
            for line in f.readlines():
                self.bookmarks.append(string.rstrip(line))
            f.close()
        except IOError:
            try:
                f = open(os.path.join(self.basedir, "stations.txt"), "r")
                for line in f.readlines():
                    self.bookmarks.append(string.rstrip(line))
                f.close()
            except IOError:
                pass

        # Minor prettification
        if bind_address == "127.0.0.1" or bind_address == "0.0.0.0":
            bind_address = "localhost"

        print "To tune in, point your browser to:"
        print "  http://" + bind_address + ":" + str(port) + "/"

        self.proxy_ready = True
        runningthreads = []

        try:
            while not self.quit:
                (clientsocket, address) = s.accept()

                if self.lastfm.debug:
                    sys.stderr.write("\nGot connection from " + repr(address) + "\n")

                t = threading.Thread(target=self.gotconnection,args=(clientsocket,))
                t.start()
                runningthreads.append(t)
        except KeyboardInterrupt:
            self.quit = 1

        print "Shutting down..."
        for t in runningthreads:
            t.join()
        print "Done! Bye!"

        s.close()
        return

if __name__ == "__main__":
    p = proxy(config.username, config.password)
    p.basedir = os.path.dirname(sys.argv[0])
    p.run(config.bind_address, config.listenport)




# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
import os, gtk
from xl import common
from gettext import gettext as _

def name(file):
    return os.path.basename(file.replace('.pyc', '.py'))

class Plugin(object):
    """
        Base object for plugins
    """
    def __init__(self):
        """
            Initializes the plugin
        """
        pass

    def initialize(self):
        pass

    def destroy(self):
        pass

class SignalContainer(object):
    def __init__(self):
        self.dict = {}

    def connect(self, obj, signal, func, *args):
        try:
            conns = self.dict[obj]
        except KeyError:
            conns = self.dict[obj] = []
        conns.append(obj.connect(signal, func, *args))

    def disconnect_object(self, obj):
        conns = self.dict.get(obj, None)
        if conns:
            for conn in conns:
                obj.disconnect(conn)
            del self.dict[obj]

    def disconnect_all(self):
        for obj, conns in self.dict.iteritems():
            for conn in conns:
                obj.disconnect(conn)
        self.dict = {}

class PluginConfigDialog(gtk.Dialog):
    """
        A generic plugin configuration dialog
    """
    def __init__(self, parent, plugin):
        """
            Initializes the dialog
        """
        # TRANSLATORS: The window title for the generic configuration dialog of plugins
        gtk.Dialog.__init__(self, _("Plugin Configuration"), parent)
        self.add_button("gtk-cancel", gtk.RESPONSE_CANCEL)
        self.add_button("gtk-ok", gtk.RESPONSE_OK)

        self.main = self.child
        self.main.set_spacing(3)

        # TRANSLATORS: Short text stating which plugin a configuration dialog is for
        label = gtk.Label(_("<b>%s Configuration</b>") %
            common.escape_xml(plugin))
        label.set_markup(label.get_label())
        self.main.pack_start(label, False, False)

class PluginInitException(Exception):
    """
        Called when a plugin can't be enabled
    """
    def __init__(self, message):
        Exception.__init__(self, message)

class DeviceDriver(object):
    def get_initial_root(self, model):
        return None

    def connect(self):
        return True

    def search_tracks(self):
        return []

    def disconnect(self):
        pass

    def get_item(self):
        pass




import httpclient
import time
import string
import playlist
import sys

class lastfm:

    def __init__(self):
        self.version = "1.3.1.1"
        self.platform = "linux"
        self.host = "ws.audioscrobbler.com"
        self.port = 80
        self.info = None
        self.playlist = playlist.playlist()
        self.debug = 0

    def parselines(self, str):
        res = {}
        vars = string.split(str, "\n")
        numerrors = 0
        for v in vars:
            x = string.split(string.rstrip(v), "=", 1)
            if len(x) == 2:
                res[x[0]] = x[1]
            elif x != [""]:
                print "(urk?", x, ")"
                numerrors = numerrors + 1
            if numerrors > 5:
                print "Too many errors parsing response."
                return {}
        return res

    def connect(self, username, password):

        s = httpclient.httpclient(self.host, self.port)
        s.req("/radio/handshake.php?version=" + self.version + "&platform=" + self.platform + "&username=" + username + "&passwordmd5=" + password + "&language=en&player=lastfmproxy")

        self.info = self.parselines(s.response)


    def getplaylist(self):

        if self.debug:
            sys.stderr.write("Fetching playlist...\n")

        s = httpclient.httpclient(self.info["base_url"])
        s.req(self.info["base_path"] + "/xspf.php?sk=" + self.info["session"] + "&discovery=0&desktop=" + self.version)

        self.playlist.parse(s.response)

        # debug
        if self.debug:
            if len(self.playlist.data.tracks):
                f = open("playlist.xspf", "w")
                f.write(s.response)
                f.close()
            elif False:
                print "No playlist?? Using cached version instead..."
                f = open("playlist.xspf", "r")
                cache = f.read()
                f.close()
                self.playlist.parse(cache)
                self.playlist.pos = 0 #len(self.playlist.data.tracks) -1

        return len(self.playlist.data.tracks)

    def command(self, cmd):
        # commands = skip, love, ban, rtp, nortp

        if self.debug:
            sys.stderr.write("command " + cmd + "\n")

        s = httpclient.httpclient(self.info["base_url"], 80)
        s.req(self.info["base_path"] + "/control.php?command=" + cmd + "&session=" + self.info["session"])
        res = self.parselines(s.response)
        
        if not res.has_key("response") or res["response"] != "OK":
            sys.stderr.write("command " + cmd + " returned:" + repr(res) + "\n")
        
        return res

    def changestation(self, url):
        
        if self.debug:
            sys.stderr.write("changestation " + url + "\n")
        
        s = httpclient.httpclient(self.info["base_url"], 80)
        s.req(self.info["base_path"] + "/adjust.php?session=" + self.info["session"] + "&url=" + url)
        res = self.parselines(s.response)

        if not res.has_key("response") or res["response"] != "OK":
            sys.stderr.write("changestation " + url + " returned:" + repr(res) + "\n")
        
        return res




from xml.sax import make_parser, handler, SAXParseException
import xml.sax


class XSPFParser( handler.ContentHandler):

    def __init__(self):
        self.path = ""
        self.curURL = ""
        self.tracks = []
        self.track = {}

    def parseFile(self, fileName):
       
        try:
            parser = make_parser()
            parser.setContentHandler(self)
            parser.parse(fileName)
            return True
        except SAXParseException:
            return False

    def parseBuff(self, buff):
        buff = '<?xml version="1.0" encoding="UTF-8"?>'+"\n"+ buff
        try:
            xml.sax.parseString(buff,self)
            return True
        except SAXParseException:
            return False

    def getTracks(self):
        return self.tracks
        
    def startElement(self, name, attrs):
        self.path += "/%s" % name
        self.content = ""
        if self.path == "/playlist/trackList/track/link":
            self.rel= attrs.get('rel',"")
        else:
            self.rel=""

    def characters(self, content):
        self.content = self.content + content

    def endElement(self, name):
        if self.path == "/playlist/trackList/track":
            self.tracks.append(self.track)
            self.track={}
            self.track["playlisttitle"] = self.title
        
        if self.path == "/playlist/trackList/track/location":
            self.track["location"]=self.content
        if self.path == "/playlist/trackList/track/title":
            self.track["title"]=self.content.encode('utf8')
        if self.path == "/playlist/trackList/track/creator":
            self.track["creator"]=self.content.encode('utf8')
        if self.path == "/playlist/trackList/track/id":
            self.track["id"] = self.content
        if self.path == "/playlist/trackList/track/album":
            self.track["album"] = self.content.encode('utf8')
        if self.path == "/playlist/trackList/track/duration":
            self.track["duration"] = self.content
        if self.path == "/playlist/trackList/track/image":
            self.track["image"] = self.content
        if self.path == "/playlist/trackList/track/lastfm:trackauth":
            self.track["trackauth"] = self.content
        if self.path == "/playlist/trackList/track/lastfm:albumId":
            self.track["albumId"] = self.content
        if self.path == "/playlist/trackList/track/lastfm:artistId":
            self.track["artistId"] = self.content

        if self.path == "/playlist/trackList/track/link":
            if self.rel=="http://www.last.fm/artistpage":
                self.track["artistpage"] = self.content
                if self.rel=="http://www.last.fm/albumpage":
                    self.track["albumpage"] = self.content
            if self.rel=="http://www.last.fm/trackpage":
                self.track["trackpage"] = self.content
            if self.rel=="http://www.last.fm/buyTrackURL":
                self.track["buytrackpage"] = self.content
            if self.rel=="http://www.last.fm/buyAlbumURL":
                self.track["buyalbumpage"] = self.content
            if self.rel=="http://www.last.fm/freeTrackURL":
                self.track["freetrackpage"] = self.content

        if self.path == "/playlist/title":
            self.title = self.content

        offset = self.path.rfind("/")
        if offset >= 0:
            self.path = self.path[0:offset]


if __name__ == "__main__":
    a=XSPFParser()
    a.parseFile("playlist.xspf")
    print a.getTracks()




#!/usr/bin/env python

# Copyright (C) 2008 Darlan Cavalcante Moreira
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


import gtk
from gettext import gettext as _
import xl.plugins as plugins
 
PLUGIN_NAME        = _("Enhanced Playlist Tabs")
PLUGIN_AUTHORS     = ['Darlan Cavalcante Moreira <darcamo@gmail.com>']
PLUGIN_VERSION     = '0.1'
PLUGIN_DESCRIPTION = _(r"""Allow closing a tab with a middle mouse click, hide the close button and expand the tabs.""")

PLUGIN_ENABLED     = False
PLUGIN_ICON        = None
 

GLADE_XML          = None
APP                = None
NOTEBOOK           = None
CONFWINDOW         = None
CONNS              = plugins.SignalContainer()
HIDECLOSEBUTTON    = False
EXPANDTABS         = False


def load_data(zip):
    """
        Called by Exaile to load the data from the zipfile
    """
    global GLADE_XML_STRING
    GLADE_XML_STRING = zip.get_data('gui.glade')
    

def initialize():
    """
        Called when the plugin is enabled
    """
    global APP, NOTEBOOK, CONFWINDOW, XML_STRING, HIDECLOSEBUTTON, EXPANDTABS, GLADE_XML

    GLADE_XML = gtk.glade.xml_new_from_buffer(GLADE_XML_STRING, len(GLADE_XML_STRING))
    CONFWINDOW        = GLADE_XML.get_widget('configwindow')
    okButton          = GLADE_XML.get_widget('OkButton')
    hideCheckButton   = GLADE_XML.get_widget('HideCheckButton')
    expandCheckButton = GLADE_XML.get_widget('ExpandCheckButton')

    CONNS.connect(okButton, 'clicked', on_OkButton_clicked)
    CONNS.connect(hideCheckButton, 'toggled', on_hideCheckButton_toggle)
    CONNS.connect(expandCheckButton, 'toggled', on_expandCheckButton_toggle)

    
    NOTEBOOK = APP.playlists_nb

    # When Exaile is initialized there is no tab (they are loaded after this plugin)
    CONNS.connect(NOTEBOOK, 'page-added', on_new_tab)

    HIDECLOSEBUTTON = APP.settings.get_boolean('hideclosebutton',default=False, plugin=plugins.name(__file__))
    EXPANDTABS = APP.settings.get_boolean('expandtabs',default=False, plugin=plugins.name(__file__))
    hideCheckButton.set_active(HIDECLOSEBUTTON)
    expandCheckButton.set_active(EXPANDTABS)

    for tab in NOTEBOOK:
        on_new_tab(NOTEBOOK, tab)
        show_hide_close_button(tab)
        NOTEBOOK.child_set_property(tab,'tab-expand',EXPANDTABS)
    return True

 
def destroy():
    """
        Called when the plugin is disabled
    """
    global CONNS, HIDECLOSEBUTTON, CONFWINDOW
    CONNS.disconnect_all()

    CONFWINDOW.hide()
    CONFWINDOW.destroy()
    CONFWINDOW = None

    HIDECLOSEBUTTON = False
    for tab in NOTEBOOK:
        show_hide_close_button(tab)
        NOTEBOOK.child_set_property(tab,'tab-expand',False)


def on_new_tab(notebook=None, tab=None, number=None):
    global CONNS, NOTEBOOK, EXPANDTABS
    tablabel = notebook.get_tab_label(tab)
    CONNS.connect(tablabel, "button-press-event", close_tab, tab)

    # If CloseButton should be invisible
    show_hide_close_button(tab)

    # If tab should expand
    NOTEBOOK.child_set_property(tab,'tab-expand',EXPANDTABS)


def show_hide_close_button(tab):
    """
        Show or hide the close button in the tab according to the value of the global variable HIDECLOSEBUTTON.
    """
    global NOTEBOOK, HIDECLOSEBUTTON
    tablabel = NOTEBOOK.get_tab_label(tab)
    hbox = tablabel.get_children()[0]
    closeButton = hbox.get_children()[1]
    if HIDECLOSEBUTTON:
        closeButton.hide()
    else:
        closeButton.show()


def close_tab(notebook=None,event=None, tab=None):
    global APP
    if (event.button == 2):
        APP.close_page(tab)


def on_OkButton_clicked(button=None):
    CONFWINDOW.hide()


def on_hideCheckButton_toggle(button=None):
    global HIDECLOSEBUTTON, NOTEBOOK
    HIDECLOSEBUTTON = button.get_active()
    APP.settings.set_boolean('hideclosebutton', HIDECLOSEBUTTON, plugin=plugins.name(__file__))
    for tab in NOTEBOOK:
        show_hide_close_button(tab)


def on_expandCheckButton_toggle(button=None):
    global EXPANDTABS
    EXPANDTABS = button.get_active()
    APP.settings.set_boolean('expandtabs', EXPANDTABS, plugin=plugins.name(__file__))
    for tab in NOTEBOOK:
        NOTEBOOK.child_set_property(tab,'tab-expand',EXPANDTABS)
    

def configure():
    """
        Called when the user clicks Configure in the Plugin Manager
    """
    global CONFWINDOW
    if(CONFWINDOW):
        CONFWINDOW.show()

